#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "binary_tree.h"
#include "queue.h"


// ------------------------------------------------------------
// ----------          data structure                ----------
// ------------------------------------------------------------

struct _Tree
{
    struct _Node* root;
};


struct _Node
{
    struct _Node* left;
    struct _Node* right;
    int value;
};


// ------------------------------------------------------------
// ----------        hidden implementation           ----------
// ------------------------------------------------------------

Node* initNode( const int value )
{
    Node* node = (Node*) malloc( sizeof( Node ) );
    if( node == NULL )
    {
        fprintf( stderr, "ERROR: There is not enough free memory available!\n" );
        exit( EXIT_FAILURE );
    }

    node->left = NULL;
    node->right = NULL;
    node->value = value;
    return node;
}


void queueBreadthFirst( Queue* queue, Node* node )
{
    // init. queue
    Queue* tmpQ = initQueue();
    push( tmpQ, node );

    // begin breadth-first search
    Node* tmp = pop( tmpQ );
    while( tmp != NULL )
    {
        // push the 'return values'
        push( queue, tmp );

        // save the acutal nodes successors
        if( tmp->left != NULL )
        {
            push( tmpQ, tmp->left );
        }
        if( tmp->right != NULL )
        {
            push( tmpQ, tmp->right );
        }

        // set next node to continue
        tmp = pop( tmpQ );
    }

    // free resources
    deleteQueue( tmpQ );
}


void queueDepthFirst( Queue* queue, Node* node )
{
    if( node != NULL )
    {
        push( queue, node );
        queueDepthFirst( queue, node->left );
        queueDepthFirst( queue, node->right );
    }
}

void inOrder( Queue* queue, Node* node )
{
    if( node->left != NULL )
    {
        inOrder( queue, node->left );
    }

    push( queue, node );

    if( node->right != NULL )
    {
        inOrder( queue, node->right );
    }
}

int max( int a, int b )
{
    return a < b ? b : a;
}


size_t treeHeightRecursive( const Node* node )
{
    if( node == NULL )
    {
        return 0;
    }
    else
    {
        return 1 + max( treeHeightRecursive( node->left ),
                        treeHeightRecursive( node->right ) );
    }
}


size_t treeSizeRecursive( const Node* node )
{
    if( node == NULL )
    {
        return 0;
    }

    size_t size = 1;
    size += treeSizeRecursive( node->left );
    size += treeSizeRecursive( node->right );

    return size;
}


Node* searchRecursive( Node* node, const int value )
{
    if( node == NULL )
    {
        return NULL;
    }

    // Is the value equal, smaller or larger than the actual value?
    if( node->value == value )
    {
        return node;
    }
    else if( node->left != NULL
             && value < node->value )
    {
        return searchRecursive( node->left, value );
    }
    else if( node->right != NULL
             && value > node->value )
    {
        return searchRecursive( node->right, value );
    }
    else // if value was not found: return some 'default value'
    {
        return NULL;
    }
}


void insertRecursive( Node** node, Node* new )
{
    if( new == NULL )
    {
        return;
    }

    // insert value in empty root
    if( *node == NULL )
    {
        *node = new;
    }
    // search in left subtree for the correct node for the value
    else if( new->value < (*node)->value )
    {
        if( (*node)->left != NULL )
        {
            insertRecursive( &((*node)->left), new );
        }
        else
        {
            (*node)->left = new;
        }
    }
    // search in right subtree for the correct node for the value
    else if( new->value > (*node)->value )
    {
        if( (*node)->right != NULL )
        {
            insertRecursive( &(*node)->right, new );
        }
        else
        {
            (*node)->right = new;
        }
    }
    // otherwise the value is already contained in the tree
}


// the only method that may changes the value of a pointer
void clearRecursive( Node** node )
{
    if( *node == NULL )
    {
        return;
    }
    else
    {
        clearRecursive( &( (*node)->left ) );
        clearRecursive( &( (*node)->right ) );
        free( *node );
        *node = NULL; // free does not change the pointer value; set it manually to NULL
    }
}


// ------------------------------------------------------------
// ----------    implementation of the interface     ----------
// ------------------------------------------------------------

Tree* initTree( void )
{
    Tree* tree = (Tree*) malloc( sizeof( Tree ) );
    if( tree == NULL )
    {
        fprintf( stderr, "ERROR: There is not enough free memory availible!\n" );
        exit( EXIT_FAILURE );
    }

    tree->root = NULL;
    return tree;
}


size_t treeHeight( const Tree* tree )
{
    return treeHeightRecursive( tree->root );
}


size_t treeSize( const Tree* tree )
{
    if( tree == NULL )
    {
        return 0;
    }

    return treeSizeRecursive( tree->root );
}


Node* search( const Tree* tree, const int value )
{
    return searchRecursive( tree->root, value );
}


void insert( Tree* tree, const int value )
{
    if( search( tree, value ) == NULL )
    {
        insertRecursive( &(tree->root), initNode( value ) );
    }
}


void erase( Tree* tree, const int value )
{
    // Strategy: Always replace the node that has to be erased by its left subtree.
    //           Then reinsert its right subtree. Our own implementation of
    //           'insertRecursive' guarantees that this works and NULL pointers
    //           are treated correctly.

    // Do we really have to do something?
    Node* node = search( tree, value );
    if( node != NULL )
    {
        // Is value the only value in the tree?
        if( node == tree->root )
        {
            tree->root = node->left;
            insertRecursive( &(tree->root), node->right );
            free( node );
        }
        else
        {
            // find parent of the node that will be erased
            Node* parent = tree->root;
            while( ( parent->left == NULL || parent->left->value != value )
                   && ( parent->right == NULL || parent->right->value != value ) )
            {
                if( value < parent->value )
                {
                    parent = parent->left;
                }
                else if( value > parent->value )
                {
                    parent = parent->right;
                }
            }

            // set parent always to node-left
            if( parent->left != NULL &&  parent->left->value == value )
            {
                parent->left = node->left;
            }
            else if( parent->right != NULL && parent->right->value == value )
            {
                parent->right = node->left;
            }

            insertRecursive( &parent, node->right );

            free( node );
        }
    }
}


void deleteTree( Tree* tree )
{
    if( tree != NULL )
    {
        clearRecursive( &(tree->root) );
    }
    tree->root = NULL;
    free( tree );
}


void printTree( const Tree* tree, const enum ORDER order )
{
    // NOTE: Since the compiler only knows the forward declaration of the
    // struct '_Queue' the 'malloc( sizeof( Queue ) )' method will not
    // work here. So, the hidden struct definition causes the need for
    // an init. method.
    Queue* queue = initQueue();
    if( order == DEPTH )
    {
        queueDepthFirst( queue, tree->root );
    }
    else if( order == BREADTH )
    {
        queueBreadthFirst( queue, tree->root );
    }
    else if( order == INORDER )
    {
	inOrder( queue, tree->root );
    }
    else
    {
        deleteQueue( queue );
        return;
    }

    printf( "values in the tree: " );
    Node* next = pop( queue );
    while( next != NULL )
    {
        printf( "%d ", next->value );
        next = pop( queue );
    }
    printf( "\n" );
    deleteQueue( queue );
}
